// Fill out your copyright notice in the Description page of Project Settings.


#include "Shuttle.h"
#include "Components/StaticMeshComponent.h"
#include "Asteroid.h"
#include "ShuttleWayPoint.h"
#include "Wayfinding.h"
#include <DrawDebugHelpers.h>
#include <Kismet/GameplayStatics.h>
#include <Kismet/KismetMathLibrary.h>
#include <Math/Vector.h>
//#include "Components/PrimitiveComponent.h"

// Sets default values
AShuttle::AShuttle()
{
 	// Set this pawn to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	// Components.
	ShuttleMesh = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("ShuttleMesh"));
	//SetRootComponent(ShuttleMesh);
	ShuttleMesh->SetupAttachment(RootComponent);

	/* Control parameters */
	// Location PID default parameters
	PitchPIDController.P = 0.1f;
	PitchPIDController.I = 0.0f;
	PitchPIDController.D = 0.001f;
	PitchPIDController.MaxOutAbs = 1.0f;

	YawPIDController.P = 0.1f;
	YawPIDController.I = 0.0f;
	YawPIDController.D = 0.001f;
	YawPIDController.MaxOutAbs = 1.0f;

	// Set shuttle initial state.
	bool _AutoPilot = false;
	float _YawControl = 0.0f;
	float _PitchControl = 0.0f;
	float _ThrustControl = 0.0f;
}

// Called when the game starts or when spawned
void AShuttle::BeginPlay()
{
	Super::BeginPlay();

	// Set initial shuttle location.
	SetActorLocation(FVector(-10000.0f, 0.0f, 15000.0f));
	SetActorRotation(FRotator(0.0f, 0.0f, 0.0f));

	// Init PID -- update function will be optimized if I and/or D are 0
	PitchPIDController.Init();
	YawPIDController.Init();

	// Configure collision callback.
	//ShuttleMesh->SetCollisionProfileName(TEXT("Overlap"));
	ShuttleMesh->OnComponentBeginOverlap.AddDynamic(this, &AShuttle::OnMeshBeginOverlap);

	// Get Wayfinding Object.
	_ShuttleWayfinder =  new Wayfinding(FVector(200000.0f,0.0f,10000.0f), GetWorld());

	_AsteroidHitCounter = 0;
}

// Called every frame
void AShuttle::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	UWorld* World = GetWorld();
	FVector AsteroidLocation;
	FVector WayPointLocation;
	FTransform ShuttleTransform = GetActorTransform();
	FVector ShuttleLocation = GetActorLocation();
	FVector ShuttleForwardVector = GetActorForwardVector();
	FVector ShuttleRightVector = GetActorRightVector();
	FRotator ShuttleRotation = GetActorRotation();
	FVector ShuttleUpVector = GetActorUpVector();
	FRotator ToWayPointRotation, RotationError;
	AAsteroid* CurrentAsteroid;
	FVector CurrentAsteroidLocation;
	unsigned char roll = 0;

	TArray<AActor*> AsteroidActors;
	UGameplayStatics::GetAllActorsOfClass(World, AAsteroid::StaticClass(), AsteroidActors);
	for (AActor* Asteroid : AsteroidActors)
	{
		FVector AsteroidVelocity = (Cast<AAsteroid>(Asteroid))->GetAsteroidVelocity();
		AsteroidLocation = Asteroid->GetActorLocation();
		//DrawDebugLine(World, ShuttleLocation, AsteroidLocation, FColor::Yellow);
		//DrawDebugSphere(World, AsteroidLocation, 1000.f, 16, FColor::Yellow);
		//DrawDebugLine(World, AsteroidLocation, AsteroidLocation + 10.0f * AsteroidVelocity, FColor::Cyan);
		//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Cyan, FString::Printf(TEXT("Asteroid Velocity: %s"),*(AsteroidVelocity).ToString()));
	}

	if (GEngine) {
		//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Shuttle Rotator: %s"), *ShuttleRotation.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("WayFinding Complete: %d"), _ShuttleWayfinder->WayFindingComplete()));
		//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Waypoint Location: %s"), *_ShuttleWayfinder->GetCurrentWaypointLocation().ToString()));
	}

	// Shuttle Way-finding.
	if (!_ShuttleWayfinder->WayFindingComplete())
	{
		//WayPointLocation = WayPoint->GetActorLocation();
		WayPointLocation = _ShuttleWayfinder->GetCurrentWaypointLocation();
		ToWayPointRotation = UKismetMathLibrary::FindLookAtRotation(ShuttleLocation, WayPointLocation);
		RotationError = ToWayPointRotation - ShuttleRotation; 
		RotationError.Normalize(); // Get shuttle-relative, not world angles.

		// Update WayFinder.
		roll = _ShuttleWayfinder->Update(this, World);

		// Control Action.
		_PitchControl = PitchPIDController.Update(RotationError.Pitch, DeltaTime);
		_YawControl = YawPIDController.Update(RotationError.Yaw, DeltaTime);
		_ThrustControl = 0.0f;

		DrawDebugLine(World, ShuttleLocation, WayPointLocation, FColor::Green);
		//DrawDebugSphere(World, WayPointLocation, 1000.f, 16, FColor::Green);

		// Highlight asteroid to avoid.
		CurrentAsteroid = _ShuttleWayfinder->GetCurrentAsteroid();
		if (CurrentAsteroid) {
			CurrentAsteroidLocation = CurrentAsteroid->GetActorLocation();
			DrawDebugSphere(World, CurrentAsteroidLocation, 1000.f, 16, FColor::Red);
			//DrawDebugLine(World, CurrentAsteroidLocation, CurrentAsteroidLocation + 10000.0f * CurrentAsteroid->GetActorForwardVector(), FColor::Red);
		}
		/*
		if (GEngine) {
			//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Waypoint Location: %s"), *WayPointLocation.ToString()));
			//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Waypoint Rotation: %s"), *ToWayPointRotation.ToString()));
			//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Waypoint Rotation Error: %s"), *RotationError.ToString()));
			//GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Pitch Control:%f Yaw Control:%f"), _PitchControl, _YawControl));
		}
		*/
		// Apply Control.
		if (_AutoPilot) 
		{
			ShuttleAngularVelocity.Yaw = _YawControl * 100.0f;
			ShuttleAngularVelocity.Pitch = _PitchControl * 100.0f;
			ShuttleVelocity.X = 1.0f;
			// Shuttle Roll Motion.
			if (roll & 0x03) { ShuttleAngularVelocity.Roll = -100.0f;}
			else if (roll & 0x01) { ShuttleAngularVelocity.Roll = -100.0f;}
			else if (roll & 0x02) { ShuttleAngularVelocity.Roll = 100.0f; }
			else { ShuttleAngularVelocity.Roll = -0.0f; }
		}
	}
	else {
		_AutoPilot = false;
	}

	// Shuttle Pitch and Yaw Motion.
	if (!ShuttleAngularVelocity.IsZero())
	{
		FRotator NewOrientation = ShuttleRotation + (ShuttleAngularVelocity * DeltaTime);
		SetActorRotation(NewOrientation);
	}
	 
	// Shuttle Linear Motion.
	if (!ShuttleVelocity.IsZero())
	{
		//FVector NewLocation = ShuttleLocation + (ShuttleVelocity * DeltaTime);
		ShuttleForwardVector *= ShuttleVelocity.X; // Thrust axis.
		//ShuttleForwardVector.Y += ShuttleVelocity.Y; // Strafe axis.
		FVector NewLocation = ShuttleLocation + ShuttleForwardVector * (10000.0f * DeltaTime);
		SetActorLocation(NewLocation);
		if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Location-> X=%f Y=%f Z=%f"), NewLocation.X, NewLocation.Y, NewLocation.Z)); }
	}

	// Useful messages.
	GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Red, FString::Printf(TEXT("Asteroid Hit Counter:%i"), _AsteroidHitCounter));
}

// Called to bind functionality to input
void AShuttle::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
	Super::SetupPlayerInputComponent(PlayerInputComponent);

	// Axis Mappings ---------------------------------------------------------------------
	// Mouse X (Yaw).
	PlayerInputComponent->BindAxis(FName("MyShuttleYaw"), this, &AShuttle::ShuttleYawMovement);
	// Mouse Y (Pitch).
	PlayerInputComponent->BindAxis(FName("MyShuttlePitch"), this, &AShuttle::ShuttlePitchMovement);
	// W & S (Thrust).
	PlayerInputComponent->BindAxis(FName("MyShuttleThrust"), this, &AShuttle::ShuttleThrustMovement);
	// A & D (Strafe).
	PlayerInputComponent->BindAxis(FName("MyShuttleStrafe"), this, &AShuttle::ShuttleStrafeMovement);
	// Q & E (Roll).
	PlayerInputComponent->BindAxis(FName("MyShuttleRoll"), this, &AShuttle::ShuttleRollMovement);

	// Action Mappings --------------------------------------------------------------------
	// H (Autopilot).
	PlayerInputComponent->BindAction(FName("MyShuttleAutoPilot"), IE_Pressed, this, &AShuttle::ToggleAutoPilot);

}

bool AShuttle::GetAutoPilot()
{
	return _AutoPilot;
}

float AShuttle::GetYawControl()
{
	return _YawControl;
}

float AShuttle::GetPitchControl()
{
	return _PitchControl;
}

float AShuttle::GetThrustControl()
{
	return _ThrustControl;
}

void AShuttle::OnMeshBeginOverlap(UPrimitiveComponent* OverlappedComponent,
									AActor* OtherActor,
									UPrimitiveComponent* OtherComp,
									int32 OtherBodyIndex,
									bool bFromSweep,
									const FHitResult& SweepResult)
{
	AAsteroid* asteroid = Cast<AAsteroid>(OtherActor);
	AShuttleWayPoint* waypoint = Cast<AShuttleWayPoint>(OtherActor);

	if (asteroid != nullptr)
	{
		if (GEngine) {
			++_AsteroidHitCounter;
		}
	}
	else if (waypoint != nullptr)
	{
		if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 3.0, FColor::Green, FString::Printf(TEXT("Hit Waypoint"))); }
		_ShuttleWayfinder->OnWaypointCollision(waypoint);
		OtherActor->Destroy();
	}

	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 3.0, FColor::Red, FString::Printf(TEXT("Collision"))); }
}

void AShuttle::ToggleAutoPilot()
{
	// Toggle AutoPilot variable.
	if (_AutoPilot)
		_AutoPilot = false;
	else
		_AutoPilot = true;

	// Debug.
	if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 3.0, FColor::Green, FString::Printf(TEXT("AutoPilot Toggled"))); }
}

void AShuttle::ShuttleYawMovement(float AxisValue)
{
	ShuttleAngularVelocity.Yaw = FMath::Clamp(AxisValue, -1.0f, 1.0f) * 100.0f;
	// Debug.
	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Yaw Value:%f"), AxisValue)); }
}

void AShuttle::ShuttlePitchMovement(float AxisValue)
{
	ShuttleAngularVelocity.Pitch = FMath::Clamp(AxisValue, -1.0f, 1.0f) * 100.0f;
	// Debug.
	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Pitch Value:%f"), AxisValue)); }
}

void AShuttle::ShuttleThrustMovement(float AxisValue)
{
	ShuttleVelocity.X = FMath::Clamp(AxisValue, -1.0f, 1.0f);
	//UE_LOG(LogTemp, Log, TEXT("Thrust Axis Value: %f"), AxisValue);
	// Debug.
	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Thrust Value:%f"), AxisValue)); }
}

void AShuttle::ShuttleStrafeMovement(float AxisValue)
{
	ShuttleVelocity.Y = FMath::Clamp(AxisValue, -1.0f, 1.0f);
	// Debug.
	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Strafe Value:%f"), AxisValue)); }
}

void AShuttle::ShuttleRollMovement(float AxisValue)
{
	ShuttleAngularVelocity.Roll = FMath::Clamp(AxisValue, -1.0f, 1.0f) * 100.0f;
	// Debug.
	//if (GEngine) { GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Green, FString::Printf(TEXT("Roll Value:%f"), AxisValue)); }
}