 // Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Pawn.h"
#include "PIDController.h"
#include "Shuttle.generated.h"

UCLASS(Blueprintable)
class NM_SCC_TEAMUNREAL_API AShuttle : public APawn
{
	GENERATED_BODY()

public:
	// Sets default values for this pawn's properties
	AShuttle();

	// Mesh Component.
	UPROPERTY(EditAnywhere, Category = "C++ Components")
	class UStaticMeshComponent* ShuttleMesh;

	// Control
	UPROPERTY(EditAnywhere, Category = "Closed-Loop Controller|Pitch PID Controller")
	FPIDController PitchPIDController;

	UPROPERTY(EditAnywhere, Category = "Closed-Loop Controller|Yaw PID Controller")
	FPIDController YawPIDController;

	// Shuttle velocity vector.
	FVector ShuttleVelocity;
	FRotator ShuttleAngularVelocity;

protected:
	bool _AutoPilot;
	float _YawControl;
	float _PitchControl;
	float _ThrustControl;
	class Wayfinding* _ShuttleWayfinder;
	int _AsteroidHitCounter;

protected:
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;

public:	
	// Called every frame
	virtual void Tick(float DeltaTime) override;

	// Called to bind functionality to input
	virtual void SetupPlayerInputComponent(class UInputComponent* PlayerInputComponent) override;

	UFUNCTION(BlueprintCallable, Blueprintpure)
	bool GetAutoPilot();

	UFUNCTION(BlueprintCallable, Blueprintpure)
	float GetYawControl();
	
	UFUNCTION(BlueprintCallable, Blueprintpure)
	float GetPitchControl();

	UFUNCTION(BlueprintCallable, Blueprintpure)
	float GetThrustControl();

	UFUNCTION()
	void OnMeshBeginOverlap(UPrimitiveComponent* OverlappedComponent,
							AActor* OtherActor,
							UPrimitiveComponent* OtherComp,
							int32 OtherBodyIndex,
							bool bFromSweep,
							const FHitResult& SweepResult);

	// Action callbacks.
	void ToggleAutoPilot();

	// Axis callbacks.
	void ShuttleYawMovement(float AxisValue);
	void ShuttlePitchMovement(float AxisValue);
	void ShuttleThrustMovement(float AxisValue);
	void ShuttleStrafeMovement(float AxisValue);
	void ShuttleRollMovement(float AxisValue);
};
