// Fill out your copyright notice in the Description page of Project Settings.


#include "Wayfinding.h"
#include "Asteroid.h"
//#include "Engine/World.h"
#include "Shuttle.h"
#include "ShuttleWayPoint.h"

Wayfinding::Wayfinding(FVector EndLocation, UWorld* World)
{
	SafetyMargin = 600.0f;
	ShuttleRadius = 1000.0f;
	_FinalDestination = SpawnWaypoint(EndLocation, nullptr, World);
	_IntermediateWaypoint = nullptr;
	_CurrentAsteroid = nullptr;
	_LastAsteroid = nullptr;
	_WaypointCount = 0;
}

unsigned char Wayfinding::Update(const class AShuttle *ShuttlePawn, UWorld* World)
{
	AActor* LineTraceHitActor;
	AAsteroid* LineTraceHitAsteroid;

	FTransform ShuttleTransform = ShuttlePawn->GetActorTransform();
	FVector ShuttleUpVector = ShuttlePawn->GetActorUpVector();
	FVector ShuttleForwardVector = ShuttlePawn->GetActorForwardVector();
	FVector ShuttleRightVector = ShuttlePawn->GetActorRightVector();

	FVector ShuttleLocation, CurrentWaypointLocation, AvoidanceWaypointLocation, HitLocation, CenterVectorEnd, WingVectorStart, WingVectorEnd;
	FHitResult hitResult;
	FCollisionObjectQueryParams objCollisionQueryParams;
	unsigned char roll = 0;

	// Use LineTrace to find potential Asteroid hit inbetween Shuttle location and waypoint location. "Unreal Engine 5 Game Development With C++ Scripting" p.232
	ShuttleLocation = ShuttleTransform.GetLocation();
	CurrentWaypointLocation = GetCurrentWaypointLocation();
	objCollisionQueryParams.AddObjectTypesToQuery(ECollisionChannel::ECC_PhysicsBody);

	// Left Side Hit?
	WingVectorStart = ShuttleLocation + 300 * ShuttleUpVector - 900 * ShuttleRightVector;
	WingVectorEnd = WingVectorStart + 10000 * ShuttleForwardVector;
	if (World->LineTraceSingleByObjectType(hitResult, WingVectorStart, WingVectorEnd, objCollisionQueryParams))
	{
		LineTraceHitActor = hitResult.GetActor();
		LineTraceHitAsteroid = Cast<AAsteroid>(hitResult.GetActor());
		// Asteroid in shuttle's path to waypoint?
		if (LineTraceHitActor != nullptr)
		{
			// Highlight hit Location.
			HitLocation = LineTraceHitActor->GetActorLocation();
			// No existing Waypoint? Spawn an Avoidance waypoint.
			if (_IntermediateWaypoint == nullptr) {
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, 1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
			// New Hit? Replace Avoidance waypoint.
			else if (_IntermediateWaypoint->GetAsteroidToAvoid() != LineTraceHitActor)
			{
				_IntermediateWaypoint->DestroyProcess();
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, 1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
		}
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Red);
		roll = 0x01;
	}
	else {
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Green);
	}
	WingVectorStart = ShuttleLocation - 300 * ShuttleUpVector - 900 * ShuttleRightVector;
	WingVectorEnd = WingVectorStart + 10000 * ShuttleForwardVector;
	if (World->LineTraceSingleByObjectType(hitResult, WingVectorStart, WingVectorEnd, objCollisionQueryParams))
	{
		LineTraceHitActor = hitResult.GetActor();
		LineTraceHitAsteroid = Cast<AAsteroid>(hitResult.GetActor());
		// Asteroid in shuttle's path to waypoint?
		if (LineTraceHitAsteroid != nullptr)
		{
			// Highlight hit Location.
			HitLocation = LineTraceHitActor->GetActorLocation();
			// No existing Waypoint? Spawn an Avoidance waypoint.
			if (_IntermediateWaypoint == nullptr) {
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, 1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
			// New Hit? Replace Avoidance waypoint.
			else if (_IntermediateWaypoint->GetAsteroidToAvoid() != LineTraceHitActor)
			{
				_IntermediateWaypoint->DestroyProcess();
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, 1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
		}
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Red);
		roll = 0x01;
	}
	else {
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Green);
	}

	// Right Side Hit?
	WingVectorStart = ShuttleLocation + 300 * ShuttleUpVector + 900 * ShuttleRightVector;
	WingVectorEnd = WingVectorStart + 10000 * ShuttleForwardVector;
	if (World->LineTraceSingleByObjectType(hitResult, WingVectorStart, WingVectorEnd, objCollisionQueryParams))
	{
		LineTraceHitActor = hitResult.GetActor();
		LineTraceHitAsteroid = Cast<AAsteroid>(hitResult.GetActor());
		// Asteroid in shuttle's path to waypoint?
		if (LineTraceHitAsteroid != nullptr)
		{
			// Highlight hit Location.
			HitLocation = LineTraceHitActor->GetActorLocation();
			// No existing Waypoint? Spawn an Avoidance waypoint.
			if (_IntermediateWaypoint == nullptr) {
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, -1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
			// New Hit? Replace Avoidance waypoint.
			else if (_IntermediateWaypoint->GetAsteroidToAvoid() != LineTraceHitActor)
			{
				_IntermediateWaypoint->DestroyProcess();
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, -1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
		}
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Red);
		roll = 0x02;
	}
	else {
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Green);
	}
	WingVectorStart = ShuttleLocation - 300 * ShuttleUpVector + 900 * ShuttleRightVector;
	WingVectorEnd = WingVectorStart + 10000 * ShuttleForwardVector;
	if (World->LineTraceSingleByObjectType(hitResult, WingVectorStart, WingVectorEnd, objCollisionQueryParams))
	{
		LineTraceHitActor = hitResult.GetActor();
		LineTraceHitAsteroid = Cast<AAsteroid>(hitResult.GetActor());
		// Asteroid in shuttle's path to waypoint?
		if (LineTraceHitAsteroid != nullptr)
		{
			// Highlight hit Location.
			HitLocation = LineTraceHitActor->GetActorLocation();
			// No existing Waypoint? Spawn an Avoidance waypoint.
			if (_IntermediateWaypoint == nullptr) {
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, -1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
			// New Hit? Replace Avoidance waypoint.
			else if (_IntermediateWaypoint->GetAsteroidToAvoid() != LineTraceHitActor)
			{
				_IntermediateWaypoint->DestroyProcess();
				AvoidanceWaypointLocation = ComputeNewWaypointLocation(HitLocation, -1.0f, ShuttlePawn, LineTraceHitAsteroid, World);
				_IntermediateWaypoint = SpawnWaypoint(AvoidanceWaypointLocation, LineTraceHitActor, World);
			}
		}
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Red);
		roll = 0x02;
	}
	else {
		DrawDebugLine(World, WingVectorStart, WingVectorEnd, FColor::Green);
	}

	// Useful Messages.
	GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Cyan, FString::Printf(TEXT("Waypoints Created: %i"), _WaypointCount));

	return roll;
}

class AShuttleWayPoint* Wayfinding::SpawnWaypoint(FVector Location, AActor* Asteroid, UWorld* World)
{
	FRotator Rotation(0.0f,0.0f,0.0f);
	FActorSpawnParameters SpawnInfo;
	AShuttleWayPoint* Waypoint = nullptr;
	//AShuttleWayPoint* Waypoint = World->SpawnActor<AShuttleWayPoint>(Location, Rotation, SpawnInfo);

	if (Asteroid != _LastAsteroid)
	{
		// Spawn the Blueprint waypoint (https://forums.unrealengine.com/t/how-to-spawn-a-blueprint-actor-via-c/78121/5).
		UObject* SpawnActor = Cast<UObject>(StaticLoadObject(UObject::StaticClass(), NULL, TEXT("/Game/Blueprints/BP_ShuttleWayPoint.BP_ShuttleWayPoint")));
		UBlueprint* GeneratedBP = Cast<UBlueprint>(SpawnActor);
		//UClass* SpawnClass = SpawnActor->StaticClass();
		Waypoint = World->SpawnActor<AShuttleWayPoint>(GeneratedBP->GeneratedClass, Location, Rotation, SpawnInfo);

		// Register avoided asteroid with spawned waypoint.
		Waypoint->AssignAsteroid(Asteroid);
		_LastAsteroid = Asteroid;

		// Update waypoint counter.
		_WaypointCount++;
	}

	// Return reference to waypoint object.
	return Waypoint;
}

FVector Wayfinding::GetCurrentWaypointLocation()
{
	// Return first waypoint in the array.
	if (_IntermediateWaypoint != nullptr)
	{
		return _IntermediateWaypoint->GetActorLocation();
	}
	/*
	if (!_WaypointQueue.IsEmpty()) 
	{
		return _WaypointQueue[0]->GetActorLocation();
	}
	*/
	else if(_FinalDestination != nullptr)
	{
		return _FinalDestination->GetActorLocation();
	}
	else
	{
		return FVector(0.0f,0.0f,0.0f);
	}
}

void Wayfinding::OnWaypointCollision(AShuttleWayPoint *Waypoint)
{
	_CurrentAsteroid = nullptr;

	if (Waypoint == _FinalDestination)
	{
		_FinalDestination = nullptr;
		//Waypoint->Destroy();
	}
	else if (Waypoint == _IntermediateWaypoint)
	{
		_IntermediateWaypoint->DestroyProcess();
		_IntermediateWaypoint = nullptr;
	}
	else
	{
		// Remove reference to waypoint.
		//AShuttleWayPoint* WaypointRef = _WaypointQueue.Pop();
	}
}

bool Wayfinding::WayFindingComplete()
{
	if (_FinalDestination == nullptr)
		return true;
	else
		return false;
}

class AAsteroid* Wayfinding::GetCurrentAsteroid()
{
	return _CurrentAsteroid;
}

FVector Wayfinding::ComputeNewWaypointLocation(FVector HitLocation, float VectorModifier, const class AShuttle *ShuttlePawn, AAsteroid* Asteroid, const UWorld *World)
{
	FVector AsteroidVelocity = Asteroid->GetAsteroidVelocity();
	FVector aPerpendicular, AvoidanceWaypointLocation;
	FVector ShuttleUpVector = ShuttlePawn->GetActorUpVector();
	FVector ShuttleDownVector = -1.0f * ShuttleUpVector;
	FVector ShuttleRightVector = ShuttlePawn->GetActorRightVector();
	FVector ShuttleLeftVector = -1.0f * ShuttleRightVector;
	FVector AonS_CompUp, AonS_CompRight;
	FVector ShuttleVectors[] = {ShuttleUpVector, ShuttleDownVector, ShuttleRightVector, ShuttleLeftVector};
	float Dots[4], minDot=0;
	int i, mini=0;

	// Note asteroid we're working with.
	_CurrentAsteroid = Asteroid;

	// Asteroid has zero velocity.
	if (0 && (AsteroidVelocity == FVector(0.0f,0.0f,0.0f)) /* || 1*/) {
		// Get "a" Vector. https://math.stackexchange.com/questions/511370/how-to-rotate-one-vector-about-another 
		aPerpendicular = (SafetyMargin + ShuttleRadius) * VectorModifier * ShuttleUpVector;
		AvoidanceWaypointLocation = HitLocation + aPerpendicular;
	} 
	// Asteroid has velocity.
	else {
		// Find Shuttle vector that has the most opposite relationship to Asteroid velocity vector.
		for (i = 0; i < 4; i++) {
			Dots[i] = FVector::DotProduct(AsteroidVelocity, ShuttleVectors[i]);
			if (Dots[i] < minDot) {
				minDot = Dots[i];
				mini = i;
			}
		}

		/*
		// Find component of asteroid velocity along shuttle's up and right vectors.
		AonS_CompUp = FVector::DotProduct(AsteroidVelocity, ShuttleUpVector) * ShuttleUpVector;
		AonS_CompRight = FVector::DotProduct(AsteroidVelocity, ShuttleRightVector) * ShuttleRightVector;
		aPerpendicular = (SafetyMargin + ShuttleRadius) * (-1.0f * (AonS_CompUp + AonS_CompRight));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("AsteroidVelocity: %s"), *AsteroidVelocity.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("ShuttleUpVector: %s"), *ShuttleUpVector.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("ShuttleRightVector: %s"), *ShuttleRightVector.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("AonS_CompUp: %s"), *AonS_CompUp.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("AonS_CompRight: %s"), *AonS_CompRight.ToString()));
		GEngine->AddOnScreenDebugMessage(-1, 60.0, FColor::Yellow, FString::Printf(TEXT("APerpendicular: %s"), *aPerpendicular.ToString()));
		*/

		aPerpendicular = (SafetyMargin + ShuttleRadius) * ShuttleVectors[mini];
		AvoidanceWaypointLocation = HitLocation + aPerpendicular;
		GEngine->AddOnScreenDebugMessage(-1, 0.0, FColor::Yellow, FString::Printf(TEXT("minDot: %f"), minDot));
	}

	return AvoidanceWaypointLocation;
}

FVector Wayfinding::CreateAvoidanceWaypointLocation(FVector ShuttleLocation, FVector NextDestination, FVector AsteroidCollisionLocation, float AsteroidRadius)
{
	FVector NewWaypointLocation;
	FVector ShuttleToCollisonVector;
	FVector AsteroidAvoidanceVector;
	float AvoidanceVectorLength;

	// place the waypoint some distance away from the asteroid.
	AvoidanceVectorLength = AsteroidRadius + ShuttleRadius + SafetyMargin;

	// calculate vector from shuttle to collison location.
	ShuttleToCollisonVector = AsteroidCollisionLocation - ShuttleLocation;

	// by default optimize avoidance vector based on the shuttles narrwest orrientation.
	// get shuttle orrientation in the direction of coliding with the asteroid.
	// callculate a vector pointing upward from the shuttle

	return NewWaypointLocation;
}

Wayfinding::~Wayfinding()
{
}
